@@ -49,7 +49,7 @@ typedef struct {
     long long tsubtotal;                    // time function spent excluding its children (include recursive calls)
     long long ttotal;                       // the total time that a function spent
     unsigned int builtin;                   // 0 for normal, 1 for ccall
-    unsigned int index;    
+    unsigned int index;
     _pit_children_info *children;
 } _pit; // profile_item
 
@@ -57,7 +57,7 @@ typedef struct {
     _cstack *cs;
     _htab *rec_levels;
     long id;
-    char *class_name;
+    char *name;
     long long t0;                           // profiling start CPU time
     unsigned long sched_cnt;                // how many times this thread is scheduled
 } _ctx; // context
@@ -84,44 +84,37 @@ static long long yappstarttick;
 static long long yappstoptick;
 static _ctx *prev_ctx;
 static _ctx *current_ctx;
+static PyObject *context_id_callback = NULL;
+static PyObject *context_name_callback = NULL;
 static PyObject *test_timings; // used for testing
 
 // defines
 #define UNINITIALIZED_STRING_VAL "N/A"
+
 #ifdef IS_PY3K // string formatting helper functions compatible with with both 2.x and 3.x
 #define PyStr_AS_CSTRING(s) PyBytes_AS_STRING(PyUnicode_AsUTF8String(s))
-#else
+#define PyStr_Check(s) PyUnicode_Check(s)
+#define PyStr_FromString(s) PyUnicode_FromString(s)
+#define PyStr_FromFormatV(fmt, vargs) PyUnicode_FromFormatV(fmt, vargs)
+#else // < Py3x
 #define PyStr_AS_CSTRING(s) PyString_AS_STRING(s)
+#define PyStr_Check(s) PyString_Check(s)
+#define PyStr_FromString(s) PyString_FromString(s)
+#define PyStr_FromFormatV(fmt, vargs) PyString_FromFormatV(fmt, vargs)
 #endif
 
 // forwards
 static _ctx * _profile_thread(PyThreadState *ts);
 
-static PyObject * 
+static PyObject *
 PyStr_FromFormat(const char *fmt, ...)
 {
     PyObject* ret;
     va_list vargs;
-    
-    va_start(vargs, fmt);
-#ifdef IS_PY3K
-    ret = PyUnicode_FromFormatV(fmt, vargs);
-#else
-    ret = PyString_FromFormatV(fmt, vargs);
-#endif
-    return ret;
-}
 
-static PyObject * 
-PyStr_FromString(const char *s)
-{
-    PyObject* ret;
-    
-#ifdef IS_PY3K
-    ret = PyUnicode_FromString(s);
-#else
-    ret = PyString_FromString(s);
-#endif
+    va_start(vargs, fmt);
+    ret = PyStr_FromFormatV(fmt, vargs);
+    va_end(vargs);
     return ret;
 }
 
@@ -168,7 +161,7 @@ _create_ctx(void)
         return NULL;
     ctx->sched_cnt = 0;
     ctx->id = 0;
-    ctx->class_name = NULL;
+    ctx->name = NULL;
     ctx->t0 = tickcount();
     ctx->rec_levels = htcreate(HT_RLEVEL_SIZE);
     if (!ctx->rec_levels)
@@ -177,41 +170,99 @@ _create_ctx(void)
 }
 
 char *
-_get_current_thread_class_name(void)
+_current_ctx_name(void)
 {
+    char *rc = NULL;
+    PyObject *callback_rc = NULL;
     PyObject *mthr, *cthr, *tattr1, *tattr2;
-
     mthr = cthr = tattr1 = tattr2 = NULL;
 
-    mthr = PyImport_ImportModuleNoBlock("threading"); // Requires Python 2.6.
-    if (!mthr)
-        goto err;
-    cthr = PyObject_CallMethod(mthr, "currentThread", "");
-    if (!cthr)
-        goto err;
-    tattr1 = PyObject_GetAttrString(cthr, "__class__");
-    if (!tattr1)
-        goto err;
-    tattr2 = PyObject_GetAttrString(tattr1, "__name__");
-    if (!tattr2)
-        goto err;
+    if (context_name_callback) {
+        callback_rc = PyObject_CallFunctionObjArgs(context_name_callback, NULL);
+        if (!callback_rc) {
+            PyErr_Print();
+            goto err;
+        }
+        if (!PyStr_Check(callback_rc)) {
+            yerr("context name callback returned non-string");
+            goto err;
+        }
 
-    return PyStr_AS_CSTRING(tattr2);
+        rc = PyStr_AS_CSTRING(callback_rc);
+        Py_CLEAR(callback_rc);
 
+        return rc;
+    } else {
+        mthr = PyImport_ImportModuleNoBlock("threading"); // Requires Python 2.6.
+        if (!mthr)
+            goto err;
+        cthr = PyObject_CallMethod(mthr, "currentThread", "");
+        if (!cthr)
+            goto err;
+        tattr1 = PyObject_GetAttrString(cthr, "__class__");
+        if (!tattr1)
+            goto err;
+        tattr2 = PyObject_GetAttrString(tattr1, "__name__");
+        if (!tattr2)
+            goto err;
+
+        Py_DECREF(mthr);
+        Py_DECREF(cthr);
+        Py_DECREF(tattr1);
+        Py_DECREF(tattr2);
+        return PyStr_AS_CSTRING(tattr2);
+    }
 err:
+    PyErr_Clear();
     Py_XDECREF(mthr);
     Py_XDECREF(cthr);
     Py_XDECREF(tattr1);
     Py_XDECREF(tattr2);
+    Py_XDECREF(callback_rc);
+    Py_CLEAR(context_name_callback);  /* Don't use the callback again. */
     return NULL;
 }
 
+static uintptr_t
+_current_context_id(PyThreadState *ts)
+{
+    uintptr_t rc;
+    PyObject *callback_rc;
+    if (context_id_callback) {
+        callback_rc = PyObject_CallFunctionObjArgs(context_id_callback, NULL);
+        if (!callback_rc) {
+            PyErr_Print();
+            goto error;
+        }
+        rc = (uintptr_t)PyLong_AsLong(callback_rc);
+        Py_DECREF(callback_rc);
+        if (PyErr_Occurred()) {
+            yerr("context id callback returned non-integer");
+            goto error;
+        }
+        return rc;
+    } else {
+        // Use thread_id instead of ts pointer, because when we create/delete many threads, some
+        // of them do not show up in the thread_stats, because ts pointers are recycled in the VM.
+        // Also, we do not want to delete thread stats unless clear_stats() is called explicitly.
+        // We rely on the OS to give us unique thread ids, this time.
+        // thread_id -> long
+        return (uintptr_t)ts->thread_id;
+    }
+
+error:
+    PyErr_Clear();
+    // don't use callback again
+    Py_XDECREF(context_id_callback);
+    context_id_callback = NULL;
+    return 0;
+}
+
 static _ctx *
 _thread2ctx(PyThreadState *ts)
 {
     _hitem *it;
-
-    it = hfind(contexts, (uintptr_t)ts->thread_id);
+    it = hfind(contexts, _current_context_id(ts));
     if (!it) {
         // callback functions in some circumtances, can be called before the context entry is not
         // created. (See issue 21). To prevent this problem we need to ensure the context entry for
@@ -265,30 +316,31 @@ _ccode2pit(void *cco)
         // get module name
         modname = NULL;
         mod = cfn->m_module;
-#ifdef IS_PY3K
-        if (mod && PyUnicode_Check(mod)) {      
-#else
-        if (mod && PyString_Check(mod)) {
-#endif
-            modname = PyStr_AS_CSTRING(mod);
-        } else if (mod && PyModule_Check(mod)) {
-            modname = (char *)PyModule_GetName(mod);
-        } 
-        
+
+        if (mod) {
+            if (PyStr_Check(mod)) {
+                modname = PyStr_AS_CSTRING(mod);
+            } else if (PyModule_Check(mod)) {
+                modname = (char *)PyModule_GetName(mod);
+            }
+        }
+
         if (modname == NULL) {
             PyErr_Clear();
             modname = "__builtin__";
         }
-        
+
         pit->modname = PyStr_FromString(modname);
         pit->lineno = 0;
-        
+
         // built-in method?
         if (cfn->m_self != NULL) {
             name = PyStr_FromString(cfn->m_ml->ml_name);
             if (name != NULL) {
-                PyObject *mo = _PyType_Lookup((PyTypeObject *)PyObject_Type(cfn->m_self), name);
+                PyObject *obj_type = PyObject_Type(cfn->m_self);
+                PyObject *mo = _PyType_Lookup((PyTypeObject *)obj_type, name);
                 Py_XINCREF(mo);
+                Py_XDECREF(obj_type);
                 Py_DECREF(name);
                 if (mo != NULL) {
                     pit->name = PyObject_Repr(mo);
@@ -325,7 +377,7 @@ _code2pit(PyFrameObject *fobj)
         return NULL;
 
     pit->name = NULL;
-    pit->modname = PyStr_FromString(PyStr_AS_CSTRING(cobj->co_filename));    
+    pit->modname = PyStr_FromString(PyStr_AS_CSTRING(cobj->co_filename));
     pit->lineno = cobj->co_firstlineno;
 
     PyFrame_FastToLocals(fobj);
@@ -337,10 +389,14 @@ _code2pit(PyFrameObject *fobj)
             if (locals) {
                 PyObject* self = PyDict_GetItemString(locals, "self");
                 if (self) {
-                    PyObject *as = PyObject_GetAttrString(self, "__class__");
-                    if (as) {
-                        as = PyObject_GetAttrString(as, "__name__");                    
-                        pit->name = PyStr_FromFormat("%s.%s", PyStr_AS_CSTRING(as), PyStr_AS_CSTRING(cobj->co_name));
+                    PyObject *class_obj = PyObject_GetAttrString(self, "__class__");
+                    if (class_obj) {
+                        PyObject *class_name = PyObject_GetAttrString(class_obj, "__name__");
+                        if (class_name) {
+                            pit->name = PyStr_FromFormat("%s.%s", PyStr_AS_CSTRING(class_name), PyStr_AS_CSTRING(cobj->co_name));
+                            Py_DECREF(class_name);
+                        }
+                        Py_DECREF(class_obj);
                     }
                 }
             }
@@ -359,7 +415,7 @@ static _pit *
 _get_frame(void)
 {
     _cstackitem *ci;
-    
+
     ci = shead(current_ctx->cs);
     if (!ci) {
         return NULL;
@@ -377,7 +433,7 @@ static _pit *
 _pop_frame(void)
 {
     _cstackitem *ci;
-    
+
     ci = spop(current_ctx->cs);
     if (!ci) {
         return NULL;
@@ -389,9 +445,9 @@ static _pit_children_info *
 _get_child_info(_pit *parent, _pit *current)
 {
     _pit_children_info *citem;
-    
-    citem = parent->children;    
-    while(citem) { 
+
+    citem = parent->children;
+    while(citem) {
         if (citem->index == current->index) {
             break;
         }
@@ -404,7 +460,7 @@ static _pit_children_info *
 _add_child_info(_pit *parent, _pit *current)
 {
     _pit_children_info *newci;
-  
+
     newci = ymalloc(sizeof(_pit_children_info));
     newci->index = current->index;
     newci->callcount = 0;
@@ -413,7 +469,7 @@ _add_child_info(_pit *parent, _pit *current)
     newci->tsubtotal = 0;
     newci->next = (struct _pit_children_info *)parent->children;
     parent->children = (_pit_children_info *)newci;
-    
+
     return newci;
 }
 
@@ -421,7 +477,7 @@ static long
 get_rec_level(uintptr_t key)
 {
     _hitem *it;
-    
+
     it = hfind(current_ctx->rec_levels, key);
     if (!it) {
         _log_err(1);
@@ -434,7 +490,7 @@ static int
 incr_rec_level(uintptr_t key)
 {
     _hitem *it;
-    
+
     it = hfind(current_ctx->rec_levels, key);
     if (it) {
         it->val++;
@@ -444,7 +500,7 @@ incr_rec_level(uintptr_t key)
             _log_err(2);
             return 0; // should not happen
         }
-    }    
+    }
     return 1;
 }
 
@@ -453,7 +509,7 @@ decr_rec_level(uintptr_t key)
 {
     _hitem *it;
     uintptr_t v;
-    
+
     it = hfind(current_ctx->rec_levels, key);
     if (it) {
         v = it->val--;  /*supress warning -- it is safe to cast long vs pointers*/
@@ -463,8 +519,8 @@ decr_rec_level(uintptr_t key)
         }
     } else {
         _log_err(3);
-        return 0; // should not happen 
-    }  
+        return 0; // should not happen
+    }
     return 1;
 }
 
@@ -472,40 +528,42 @@ static long long
 _get_frame_elapsed(void)
 {
     _cstackitem *ci;
-    _pit *cp;    
+    _pit *cp;
     long long result;
-    
+
     ci = shead(current_ctx->cs);
     if (!ci) {
         return 0LL;
     }
     cp = ci->ckey;
-    
-    if (test_timings) { 
-        //printf("name:%s_%d \r\n", PyStr_AS_CSTRING(cp->name), rlevel);
-        uintptr_t rlevel = get_rec_level((uintptr_t)cp);        
-        PyObject *tval = PyDict_GetItem(test_timings, 
-            PyStr_FromFormat("%s_%d", PyStr_AS_CSTRING(cp->name), rlevel));
-        if (tval) {            
+
+    if (test_timings) {
+        uintptr_t rlevel = get_rec_level((uintptr_t)cp);
+        PyObject *formatted_string = PyStr_FromFormat(
+                "%s_%d", PyStr_AS_CSTRING(cp->name), rlevel);
+
+        PyObject *tval = PyDict_GetItem(test_timings, formatted_string);
+        Py_DECREF(formatted_string);
+        if (tval) {
             result = PyLong_AsLong(tval);
         } else {
             result = DEFAULT_TEST_ELAPSED_TIME;
         }
-        
+
     } else {
         result = tickcount() - ci->t0;
     }
-    
+
     return result;
 }
 
 static void
 _call_enter(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
 {
-    _pit *cp,*pp;    
+    _pit *cp,*pp;
     _cstackitem *ci;
     _pit_children_info *pci;
-    
+
     if (ccall) {
         cp = _ccode2pit((PyCFunctionObject *)arg);
     } else {
@@ -518,19 +576,19 @@ _call_enter(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
         _log_err(4);
         return;
     }
-    
+
     // create/update children info if we have a valid parent
-    pp = _get_frame();    
+    pp = _get_frame();
     if (pp) {
-        pci = _get_child_info(pp, cp);    
+        pci = _get_child_info(pp, cp);
         if(!pci)
         {
             pci = _add_child_info(pp, cp);
-        }    
+        }
         pci->callcount++;
         incr_rec_level((uintptr_t)pci);
     }
-    
+
     ci = _push_frame(cp);
     if (!ci) { // runaway! (defensive)
         _log_err(5);
@@ -545,19 +603,19 @@ _call_enter(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
 static void
 _call_leave(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
 {
-    long long elapsed;    
-    _pit *cp, *pp, *ppp;    
+    long long elapsed;
+    _pit *cp, *pp, *ppp;
     _pit_children_info *pci,*ppci;
-            
-    elapsed = _get_frame_elapsed();    
-    
-    // leaving a frame while callstack is empty? 
+
+    elapsed = _get_frame_elapsed();
+
+    // leaving a frame while callstack is empty?
     cp = _pop_frame();
     if (!cp)
     {
         return;
     }
-    
+
     // is this the last function in the callstack?
     pp = _pop_frame();
     if (!pp) {
@@ -567,9 +625,9 @@ _call_leave(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
         decr_rec_level((uintptr_t)cp);
         return;
     }
-    
+
     // get children info
-    pci = _get_child_info(pp, cp);    
+    pci = _get_child_info(pp, cp);
     if(!pci)
     {
         _log_err(6);
@@ -577,36 +635,36 @@ _call_leave(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
     }
     // a calls b. b's elapsed time is subtracted from a's tsub and a adds its own elapsed it is leaving.
     pp->tsubtotal -= elapsed;
-    cp->tsubtotal += elapsed;    
-        
+    cp->tsubtotal += elapsed;
+
     // a calls b calls c. child c's elapsed time is subtracted from child b's tsub and child b adds its
     // own elapsed when it is leaving
     ppp = _get_frame();
     if (ppp) {
-        ppci = _get_child_info(ppp, pp);    
+        ppci = _get_child_info(ppp, pp);
         if(!ppci)
         {
             _log_err(7);
             return;
-        }            
+        }
         ppci->tsubtotal -= elapsed;
     }
     pci->tsubtotal += elapsed;
-    
-    // wait for the top-level function/parent/child to update timing values accordingly.              
+
+    // wait for the top-level function/parent/child to update timing values accordingly.
     if (get_rec_level((uintptr_t)cp) == 1) {
         cp->ttotal += elapsed;
         cp->nonrecursive_callcount++;
         pci->nonrecursive_callcount++;
     }
-    
+
     if (get_rec_level((uintptr_t)pci) == 1) {
         pci->ttotal += elapsed;
     }
-    
+
     decr_rec_level((uintptr_t)pci);
     decr_rec_level((uintptr_t)cp);
-    
+
     if (!_push_frame(pp)) {
         _log_err(8);
         return; //defensive
@@ -627,27 +685,27 @@ _yapp_callback(PyObject *self, PyFrameObject *frame, int what,
                PyObject *arg)
 {
     PyObject *last_type, *last_value, *last_tb;
-        
+
     PyErr_Fetch(&last_type, &last_value, &last_tb);
-    
+
     // get current ctx
-    current_ctx = _thread2ctx(frame->f_tstate);
+    current_ctx = _thread2ctx(PyThreadState_GET());
     if (!current_ctx) {
         _log_err(9);
         goto finally;
     }
-    
+
     // update ctx stats
     if (prev_ctx != current_ctx) {
         current_ctx->sched_cnt++;
     }
     prev_ctx = current_ctx;
-    if (!current_ctx->class_name)
+    if (!current_ctx->name)
     {
-        current_ctx->class_name = _get_current_thread_class_name();
+        current_ctx->name = _current_ctx_name();
     }
-    
-        
+
+
     switch (what) {
     case PyTrace_CALL:
         _call_enter(self, frame, arg, 0);
@@ -655,8 +713,11 @@ _yapp_callback(PyObject *self, PyFrameObject *frame, int what,
     case PyTrace_RETURN: // either normally or with an exception
         _call_leave(self, frame, arg, 0);
         break;
+    /* case PyTrace_EXCEPTION:
+        If the exception results in the function exiting, a
+        PyTrace_RETURN event will be generated, so we don't need to
+        handle it. */
 
-#ifdef PyTrace_C_CALL	// not defined in Python <= 2.3
     case PyTrace_C_CALL:
         if (PyCFunction_Check(arg))
             _call_enter(self, frame, arg, 1); // set ccall to true
@@ -667,13 +728,12 @@ _yapp_callback(PyObject *self, PyFrameObject *frame, int what,
         if (PyCFunction_Check(arg))
             _call_leave(self, frame, arg, 1);
         break;
-#endif
     default:
         break;
     }
 
     goto finally;
-    
+
 finally:
     if (last_type) {
         PyErr_Restore(last_type, last_value, last_tb);
@@ -685,6 +745,7 @@ finally:
 static _ctx *
 _profile_thread(PyThreadState *ts)
 {
+    uintptr_t ctx_id;
     _ctx *ctx;
     _hitem *it;
 
@@ -692,15 +753,11 @@ _profile_thread(PyThreadState *ts)
     if (!ctx) {
         return NULL;
     }
-    
-    // Use thread_id instead of ts pointer, because when we create/delete many threads, some
-    // of them do not show up in the thread_stats, because ts pointers are recycled in the VM.
-    // Also, we do not want to delete thread stats unless clear_stats() is called explicitly.
-    // We rely on the OS to give us unique thread ids, this time.
-    // thread_id -> long
-    it = hfind(contexts, (uintptr_t)ts->thread_id);
+
+    ctx_id = _current_context_id(ts);
+    it = hfind(contexts, ctx_id);
     if (!it) {
-        if (!hadd(contexts, (uintptr_t)ts->thread_id, (uintptr_t)ctx)) {
+        if (!hadd(contexts, ctx_id, (uintptr_t)ctx)) {
             _del_ctx(ctx);
             if (!flput(flctx, ctx)) {
                 _log_err(10);
@@ -712,8 +769,8 @@ _profile_thread(PyThreadState *ts)
 
     ts->use_tracing = 1;
     ts->c_profilefunc = _yapp_callback;
-    ctx->id = ts->thread_id;
-    
+    ctx->id = ctx_id;
+
     return ctx;
 }
 
@@ -783,11 +840,11 @@ profile_event(PyObject *self, PyObject *args)
     if (!PyArg_ParseTuple(args, "OOO", &frame, &event, &arg)) {
         return NULL;
     }
-    
+
     if (flags.multithreaded) {
         _ensure_thread_profiled(PyThreadState_GET());
     }
-    
+
     ev = PyStr_AS_CSTRING(event);
 
     if (strcmp("call", ev)==0)
@@ -801,42 +858,9 @@ profile_event(PyObject *self, PyObject *args)
     else if (strcmp("c_exception", ev)==0)
         _yapp_callback(self, frame, PyTrace_C_EXCEPTION, arg);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
-static PyObject*
-start(PyObject *self, PyObject *args)
-{
-    if (yapprunning) {
-        Py_INCREF(Py_None);
-        return Py_None;
-    }
-
-    if (!PyArg_ParseTuple(args, "ii", &flags.builtins, &flags.multithreaded))
-        return NULL;
-
-    if (!_init_profiler()) {
-        PyErr_SetString(YappiProfileError, "profiler cannot be initialized.");
-        return NULL;
-    }
-    
-    if (flags.multithreaded) {
-        _enum_threads(&_profile_thread);
-    } else {
-        _ensure_thread_profiled(PyThreadState_GET());
-    }
-    
-    yapprunning = 1;
-    yapphavestats = 1;
-    time (&yappstarttime);
-    yappstarttick = tickcount();
-    
-    Py_INCREF(Py_None);
-    return Py_None;
-}
-
-
 static long long
 _calc_cumdiff(long long a, long long b)
 {
@@ -862,36 +886,55 @@ _ctxenumdel(_hitem *item, void *arg)
     return 0;
 }
 
-static PyObject*
-stop(PyObject *self, PyObject *args)
+// start profiling. return 1 on success, or 0 and set exception.
+static int
+_start(void)
 {
-    if (!yapprunning) {
-        Py_INCREF(Py_None);
-        return Py_None;
+    if (yapprunning)
+        return 1;
+
+    if (!_init_profiler()) {
+        PyErr_SetString(YappiProfileError, "profiler cannot be initialized.");
+        return 0;
     }
 
+    if (flags.multithreaded) {
+        _enum_threads(&_profile_thread);
+    } else {
+        _ensure_thread_profiled(PyThreadState_GET());
+    }
+
+    yapprunning = 1;
+    yapphavestats = 1;
+    time (&yappstarttime);
+    yappstarttick = tickcount();
+    return 1;
+}
+
+static void
+_stop(void)
+{
+    if (!yapprunning)
+        return;
+
     _enum_threads(&_unprofile_thread);
 
     yapprunning = 0;
     yappstoptick = tickcount();
-
-    Py_INCREF(Py_None);
-    return Py_None;
 }
 
 static PyObject*
 clear_stats(PyObject *self, PyObject *args)
-{    
+{
     if (!yapphavestats) {
-        Py_INCREF(Py_None);
-        return Py_None;
+        Py_RETURN_NONE;
     }
-    
+
     henum(pits, _pitenumdel, NULL);
     htdestroy(pits);
     henum(contexts, _ctxenumdel, NULL);
     htdestroy(contexts);
-    
+
     fldestroy(flpit);
     fldestroy(flctx);
     yappinitialized = 0;
@@ -904,8 +947,7 @@ clear_stats(PyObject *self, PyObject *args)
     YMEMLEAKCHECK();
 #endif
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 // normalizes the time count if test_timing is not set.
@@ -926,9 +968,9 @@ _ctxenumstat(_hitem *item, void *arg)
     _ctx *ctx;
     long long cumdiff;
     PyObject *exc;
-    
+
     ctx = (_ctx *)item->val;
-    
+
     if(ctx->sched_cnt == 0) {
         // we return here because if sched_cnt is zero, then this means not any single function
         // executed in the context of this thread. We do not want to show any thread stats for this case especially
@@ -938,21 +980,22 @@ _ctxenumstat(_hitem *item, void *arg)
         // second invocation of test_start_flags() generates this situation.
         return 0;
     }
-    
-    tcname = ctx->class_name;
+
+    tcname = ctx->name;
     if (tcname == NULL)
         tcname = UNINITIALIZED_STRING_VAL;
     efn = (PyObject *)arg;
 
     cumdiff = _calc_cumdiff(tickcount(), ctx->t0);
-    
-    exc = PyObject_CallFunction(efn, "((skfk))", tcname, ctx->id, 
+
+    exc = PyObject_CallFunction(efn, "((skfk))", tcname, ctx->id,
         cumdiff * tickfactor(), ctx->sched_cnt);
     if (!exc) {
         PyErr_Print();
         return 1; // abort enumeration
     }
 
+    Py_DECREF(exc);
     return 0;
 }
 
@@ -962,8 +1005,7 @@ enum_thread_stats(PyObject *self, PyObject *args)
     PyObject *enumfn;
 
     if (!yapphavestats) {
-        Py_INCREF(Py_None);
-        return Py_None;
+        Py_RETURN_NONE;
     }
 
     if (!PyArg_ParseTuple(args, "O", &enumfn)) {
@@ -978,8 +1020,7 @@ enum_thread_stats(PyObject *self, PyObject *args)
 
     henum(contexts, _ctxenumstat, enumfn);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static int
@@ -991,7 +1032,9 @@ _pitenumstat(_hitem *item, void * arg)
     PyObject *children;
     _pit_children_info *pci;
 
+    children = NULL;
     pt = (_pit *)item->val;
+
     // do not show builtin pits if specified
     if  ((!flags.builtins) && (pt->builtin)) {
         return 0;
@@ -1003,12 +1046,16 @@ _pitenumstat(_hitem *item, void * arg)
     children = PyList_New(0);
     pci = pt->children;
     while(pci) {
+        PyObject *stats_tuple;
         // normalize tsubtotal. tsubtotal being negative is an expected situation.
         if (pci->tsubtotal < 0) {
             pci->tsubtotal = 0;
-        }        
-        PyList_Append(children, Py_BuildValue("Ikkff", pci->index, pci->callcount, 
-                pci->nonrecursive_callcount, _normt(pci->ttotal), _normt(pci->tsubtotal) ));
+        }
+        stats_tuple = Py_BuildValue("Ikkff", pci->index, pci->callcount,
+                pci->nonrecursive_callcount, _normt(pci->ttotal),
+                _normt(pci->tsubtotal));
+        PyList_Append(children, stats_tuple);
+        Py_DECREF(stats_tuple);
         pci = (_pit_children_info *)pci->next;
     }
     // normalize tsubtotal. tsubtotal being negative is an expected situation.
@@ -1016,23 +1063,49 @@ _pitenumstat(_hitem *item, void * arg)
         pt->tsubtotal = 0;
     }
     exc = PyObject_CallFunction(efn, "((OOkkkIffIO))", pt->name, pt->modname, pt->lineno, pt->callcount,
-                        pt->nonrecursive_callcount, pt->builtin, _normt(pt->ttotal), _normt(pt->tsubtotal), 
+                        pt->nonrecursive_callcount, pt->builtin, _normt(pt->ttotal), _normt(pt->tsubtotal),
                         pt->index, children);
     if (!exc) {
         PyErr_Print();
+        Py_XDECREF(children);
         return 1; // abort enumeration
     }
+
+    Py_DECREF(exc);
+    Py_XDECREF(children);
     return 0;
 }
 
+static PyObject*
+start(PyObject *self, PyObject *args)
+{
+    if (yapprunning)
+        Py_RETURN_NONE;
+
+    if (!PyArg_ParseTuple(args, "ii", &flags.builtins, &flags.multithreaded))
+        return NULL;
+
+    if (!_start())
+        // error
+        return NULL;
+
+    Py_RETURN_NONE;
+}
+
+static PyObject*
+stop(PyObject *self)
+{
+    _stop();
+    Py_RETURN_NONE;
+}
+
 static PyObject*
 enum_func_stats(PyObject *self, PyObject *args)
 {
     PyObject *enumfn;
 
     if (!yapphavestats) {
-        Py_INCREF(Py_None);
-        return Py_None;
+        Py_RETURN_NONE;
     }
 
     if (!PyArg_ParseTuple(args, "O", &enumfn)) {
@@ -1047,8 +1120,7 @@ enum_func_stats(PyObject *self, PyObject *args)
 
     henum(pits, _pitenumstat, enumfn);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static PyObject *
@@ -1063,22 +1135,65 @@ get_mem_usage(PyObject *self, PyObject *args)
     return Py_BuildValue("l", ymemusage());
 }
 
+static PyObject *
+set_context_id_callback(PyObject *self, PyObject *args)
+{
+    PyObject* new_callback;
+
+    if (!PyArg_ParseTuple(args, "O", &new_callback)) {
+        return NULL;
+    }
+
+    if (new_callback == Py_None) {
+        Py_CLEAR(context_id_callback);
+        Py_RETURN_NONE;
+    } else if (!PyCallable_Check(new_callback)) {
+        PyErr_SetString(PyExc_TypeError, "callback should be a function.");
+        return NULL;
+    }
+    Py_XDECREF(context_id_callback);
+    Py_INCREF(new_callback);
+    context_id_callback = new_callback;
+
+    Py_RETURN_NONE;
+}
+
+static PyObject *
+set_context_name_callback(PyObject *self, PyObject *args)
+{
+    PyObject* new_callback;
+    if (!PyArg_ParseTuple(args, "O", &new_callback)) {
+        return NULL;
+    }
+
+    if (new_callback == Py_None) {
+        Py_CLEAR(context_name_callback);
+        Py_RETURN_NONE;
+    } else if (!PyCallable_Check(new_callback)) {
+        PyErr_SetString(PyExc_TypeError, "callback should be a function.");
+        return NULL;
+    }
+    Py_XDECREF(context_name_callback);
+    Py_INCREF(new_callback);
+    context_name_callback = new_callback;
+    Py_RETURN_NONE;
+}
+
 static PyObject *
 set_test_timings(PyObject *self, PyObject *args)
 {
     if (!PyArg_ParseTuple(args, "O", &test_timings)) {
         return NULL;
     }
-    
+
     if (!PyDict_Check(test_timings))
     {
         PyErr_SetString(YappiProfileError, "timings should be dict.");
         return NULL;
     }
     Py_INCREF(test_timings);
-    
-    Py_INCREF(Py_None);
-    return Py_None;
+
+    Py_RETURN_NONE;
 }
 
 static PyObject *
@@ -1093,8 +1208,7 @@ set_clock_type(PyObject *self, PyObject *args)
     // return silently if same clock_type
     if (clock_type == get_timing_clock_type())
     {
-        Py_INCREF(Py_None);
-        return Py_None;
+        Py_RETURN_NONE;
     }
     
     if (yapphavestats) {
@@ -1107,21 +1221,27 @@ set_clock_type(PyObject *self, PyObject *args)
         return NULL;
     }   
     
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static PyObject *
-get_clock_type(PyObject *self, PyObject *args)
+get_clock_time(PyObject *self, PyObject *args)
 {
-    PyObject *type,*api,*result,*resolution;
+    return PyFloat_FromDouble(tickfactor() * tickcount());
+}
+
+static PyObject *
+get_clock_info(PyObject *self, PyObject *args)
+{
+    PyObject *api = NULL;
+    PyObject *result = NULL;
+    PyObject *resolution = NULL;
     clock_type_t clk_type;
 
     result = PyDict_New();
     
     clk_type = get_timing_clock_type();
     if (clk_type == WALL_CLOCK) {
-        type = Py_BuildValue("s", "wall");
 #if defined(_WINDOWS)
         api = Py_BuildValue("s", "queryperformancecounter");
         resolution = Py_BuildValue("s", "100ns");    
@@ -1130,7 +1250,6 @@ get_clock_type(PyObject *self, PyObject *args)
         resolution = Py_BuildValue("s", "100ns");
 #endif
     }  else {
-        type = Py_BuildValue("s", "cpu");
 #if defined(USE_CLOCK_TYPE_GETTHREADTIMES)
         api = Py_BuildValue("s", "getthreadtimes");
         resolution = Py_BuildValue("s", "100ns");
@@ -1145,42 +1264,92 @@ get_clock_type(PyObject *self, PyObject *args)
         resolution = Py_BuildValue("s", "1000ns");
 #endif
     }
-    
-    PyDict_SetItemString(result, "type", type);
+
     PyDict_SetItemString(result, "api", api);
     PyDict_SetItemString(result, "resolution", resolution);
 
+    Py_XDECREF(api);
+    Py_XDECREF(resolution);
     return result;
 }
 
+static PyObject *
+get_clock_type(PyObject *self, PyObject *args)
+{
+    clock_type_t clk_type;
+    
+    clk_type = get_timing_clock_type();
+    if (clk_type == WALL_CLOCK) {
+        return Py_BuildValue("s", "wall");
+    }  else {
+        return Py_BuildValue("s", "cpu");
+    }
+}
+
+static PyObject *
+shift_context_time(PyObject *self, PyObject *args)
+{
+    int i;
+    long context_id;
+    double amount;
+    long long shifted_amount;
+    _hitem *it;
+    _ctx *ctx;
+
+    if (!PyArg_ParseTuple(args, "ld", &context_id, &amount)) {
+        return NULL;
+    }
+
+    shifted_amount = (long long)(amount / tickfactor());
+    it = hfind(contexts, context_id);
+    if (!it || !it->val) {
+        // This context hasn't executed yet during this Yappi run; just abort.
+        Py_RETURN_NONE;
+    }
+
+     // Advance the start time for each frame in this context's call stack
+     // by the duration for which this context was paused.
+    ctx = (_ctx *)it->val;
+    for (i = 0; i <= ctx->cs->head; i++) {
+        ctx->cs->_items[i].t0 += shifted_amount;
+    }
+
+    // advance the start time for the whole context by the pause duration
+    ctx->t0 += shifted_amount;
+    Py_RETURN_NONE;
+}
+
 static PyObject*
 get_start_flags(PyObject *self, PyObject *args)
 {
-    PyObject *result;
+    PyObject *result = NULL;
+    PyObject *profile_builtins = NULL;
+    PyObject *profile_multithread = NULL;
     
     if (!yapphavestats) {
-        Py_INCREF(Py_None);
-        return Py_None;
+        Py_RETURN_NONE;
     }
 
+    profile_builtins = Py_BuildValue("i", flags.builtins);
+    profile_multithread = Py_BuildValue("i", flags.multithreaded);
     result = PyDict_New();
-    PyDict_SetItemString(result, "profile_builtins", Py_BuildValue("i", flags.builtins));
-    PyDict_SetItemString(result, "profile_multithread", Py_BuildValue("i", flags.multithreaded));
+    PyDict_SetItemString(result, "profile_builtins", profile_builtins);
+    PyDict_SetItemString(result, "profile_multithread", profile_multithread);
     
+    Py_XDECREF(profile_builtins);
+    Py_XDECREF(profile_multithread);
     return result;
 }
 
 static PyObject*
 _pause(PyObject *self, PyObject *args)
 {
-    if (yapprunning)
-    {
+    if (yapprunning) {
         paused = 1;
-        stop(self, NULL);
+        _stop();
     }
-    
-    Py_INCREF(Py_None);
-    return Py_None;
+
+    Py_RETURN_NONE;
 }
 
 static PyObject*
@@ -1189,23 +1358,29 @@ _resume(PyObject *self, PyObject *args)
     if (paused)
     {
         paused = 0;        
-        start(self, Py_BuildValue("ii", flags.builtins, flags.multithreaded));
+        if (!_start())
+            // error
+            return NULL;
     }
     
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static PyMethodDef yappi_methods[] = {
     {"start", start, METH_VARARGS, NULL},
-    {"stop", stop, METH_VARARGS, NULL},
+    {"stop", (PyCFunction)stop, METH_NOARGS, NULL},
     {"enum_func_stats", enum_func_stats, METH_VARARGS, NULL},
     {"enum_thread_stats", enum_thread_stats, METH_VARARGS, NULL},
     {"clear_stats", clear_stats, METH_VARARGS, NULL},
     {"is_running", is_running, METH_VARARGS, NULL},
     {"get_clock_type", get_clock_type, METH_VARARGS, NULL},
     {"set_clock_type", set_clock_type, METH_VARARGS, NULL},
+    {"get_clock_time", get_clock_time, METH_VARARGS, NULL},
+    {"get_clock_info", get_clock_info, METH_VARARGS, NULL},
+    {"shift_context_time", shift_context_time, METH_VARARGS, NULL},
     {"get_mem_usage", get_mem_usage, METH_VARARGS, NULL},
+    {"set_context_id_callback", set_context_id_callback, METH_VARARGS, NULL},
+    {"set_context_name_callback", set_context_name_callback, METH_VARARGS, NULL},
     {"_get_start_flags", get_start_flags, METH_VARARGS, NULL}, // for internal usage.
     {"_set_test_timings", set_test_timings, METH_VARARGS, NULL}, // for internal usage.
     {"_profile_event", profile_event, METH_VARARGS, NULL}, // for internal usage.
@@ -1216,15 +1391,15 @@ static PyMethodDef yappi_methods[] = {
 
 #ifdef IS_PY3K
 static struct PyModuleDef _yappi_module = {
-        PyModuleDef_HEAD_INIT,
-        "_yappi",
-        _yappi__doc__,
-        -1,
-        yappi_methods,
-        NULL,
-        NULL,
-        NULL,
-        NULL
+    PyModuleDef_HEAD_INIT,
+    "_yappi",
+    _yappi__doc__,
+    -1,
+    yappi_methods,
+    NULL,
+    NULL,
+    NULL,
+    NULL
 };
 #endif
 
